@@ -32,15 +32,17 @@
 * OF SUCH DAMAGE.
 ****************************************************************************/
 
-#include <pdal/compression/LazPerfVlrCompression.hpp>
-
 #include "LasHeader.hpp"
 #include "LasReader.hpp"
+#include "private/las/ChunkInfo.hpp"
 #include "private/las/Header.hpp"
 #include "private/las/Srs.hpp"
+#include "private/las/Tile.hpp"
 #include "private/las/Utils.hpp"
 #include "private/las/Vlr.hpp"
 
+#include <condition_variable>
+#include <mutex>
 #include <sstream>
 #include <string.h>
 
@@ -49,9 +51,10 @@
 #include <pdal/PointView.hpp>
 #include <pdal/QuickInfo.hpp>
 #include <pdal/util/Extractor.hpp>
-#include <pdal/util/FileUtils.hpp>
 #include <pdal/util/IStream.hpp>
 #include <pdal/util/ProgramArgs.hpp>
+#include <lazperf/readers.hpp>
+#include <arbiter/arbiter.hpp>
 
 namespace pdal
 {
@@ -59,6 +62,8 @@ namespace pdal
 namespace
 {
 
+constexpr int DefaultNumThreads = 7;
+
 struct invalid_stream : public std::runtime_error
 {
     invalid_stream(const std::string& msg) : std::runtime_error(msg)
@@ -130,9 +135,11 @@ struct LasReader::Options
     std::string compression;
     bool useEbVlr;
     StringList ignoreVLROption;
+    bool ignoreMissingVLRs;
     bool fixNames;
     PointId start;
     bool nosrs;
+    int numThreads;
     SrsOrderSpec srsVlrOrder;
 };
 
@@ -141,27 +148,38 @@ struct LasReader::Private
     Options opts;
     las::Header header;
     LasHeader apiHeader;
-    LazPerfVlrDecompressor *decompressor;
-    std::vector<char> decompressorBuf;
-    point_count_t index;
+    uint64_t index;
     las::VlrList ignoreVlrs;
     las::VlrList vlrs;
     las::Srs srs;
+    las::TilePtr currentTile;
+    las::ChunkInfo chunkInfo;
+    std::vector<las::TilePtr> tiles;
     std::vector<las::ExtraDim> extraDims;
+    ThreadPool pool;
+    // One past the Index of the last point we want to fetch.
+    PointId end;
+    // The index of the chunk we want to fetch next.
+    uint32_t nextFetchChunk;
+    // The point ID of the point we want to fetch next.
+    uint64_t nextFetchPoint;
+    // The index of the chunk (tile) we want to read data from.
+    uint32_t nextReadChunk;
+    std::function<void()> queueNext;
+    std::function<void(PointRef&, const char *, size_t)> loadPoint;
+    std::mutex mutex;
+    std::condition_variable processedCv;
+    bool isRemote;
 
-    Private() : apiHeader(header, srs, vlrs), decompressor(nullptr), index(0)
+    Private() : apiHeader(header, srs, vlrs), index(0), pool(DefaultNumThreads), isRemote(false)
     {}
 };
 
 LasReader::LasReader() : d(new Private)
 {}
 
-
 LasReader::~LasReader()
-{
-    delete d->decompressor;
-}
-
+{}
 
 void LasReader::addArgs(ProgramArgs& args)
 {
@@ -170,10 +188,13 @@ void LasReader::addArgs(ProgramArgs& args)
     args.add("compression", "Decompressor to use", d->opts.compression, "EITHER");
     args.add("use_eb_vlr", "Use extra bytes VLR for 1.0 - 1.3 files", d->opts.useEbVlr);
     args.add("ignore_vlr", "VLR userid/recordid to ignore", d->opts.ignoreVLROption );
+    args.add("ignore_missing_vlrs", "Ignore any missing VLRs rather than throwing an error",
+             d->opts.ignoreMissingVLRs);
     args.add("start", "Point at which reading should start (0-indexed).", d->opts.start);
     args.add("fix_dims", "Make invalid dimension names valid by changing "
         "invalid characters to '_'", d->opts.fixNames, true);
     args.add("nosrs", "Skip reading/processing file SRS", d->opts.nosrs);
+    args.add("threads", "Thread pool size", d->opts.numThreads, DefaultNumThreads);
     args.add("srs_vlr_order", "Preference order to read SRS VLRs",
         d->opts.srsVlrOrder);
 }
@@ -209,16 +230,34 @@ uint64_t LasReader::vlrData(const std::string& userId, uint16_t recordId, char c
     return vlr->dataVec.size();
 }
 
+// Number of points we're wanting to fetch.
 point_count_t LasReader::getNumPoints() const
 {
-    return d->header.pointCount() - d->opts.start;
+    return d->end - d->opts.start;
 }
 
 void LasReader::initialize(PointTableRef table)
 {
+    tryLoadRemote();
     initializeLocal(table, m_metadata);
 }
 
+void LasReader::tryLoadRemote()
+{
+    d->isRemote = Utils::isRemote(m_filename);
+    if (d->isRemote)
+    {
+        // Here, we're assigning the local filename to "remoteFilename".
+        // This is fixed by the swap on the next line.
+        std::string remoteFilename = Utils::tempFilename(m_filename);
+        std::swap(remoteFilename, m_filename);
+
+        // Fetch the remote file and write to the local file.
+        arbiter::Arbiter a;
+        a.put(m_filename, a.getBinary(remoteFilename));
+    }
+}
+
 QuickInfo LasReader::inspect()
 {
     QuickInfo qi;
@@ -244,18 +283,21 @@ QuickInfo LasReader::inspect()
 }
 
 
-void LasReader::createStream()
+// IMPORTANT NOTE: Unless you're going to totally overhaul things, you *must* use this
+//  virtual function to create a stream that you *must* use. NITF files contain embedded
+//  LAS files and this allows native and embedded files to be accessed identically. All
+//  the file positioning works assuming that the file is pure LAS/LAZ thanks to this.
+LasReader::LasStreamPtr LasReader::createStream()
 {
-    if (!m_streamIf)
-        m_streamIf.reset(new LasStreamIf(m_filename));
-
-    if (!m_streamIf->m_istream)
+    LasStreamPtr s(new LasStreamIf(m_filename));
+    if (!s->isOpen())
     {
         std::ostringstream oss;
         oss << "Unable to open stream for '"
-            << m_filename <<"' with error '" << strerror(errno) <<"'";
+            << m_filename << "' with error '" << strerror(errno) << "'";
         throw pdal_error(oss.str());
     }
+    return s;
 }
 
 
@@ -275,14 +317,15 @@ void LasReader::initializeLocal(PointTableRef table, MetadataNode& m)
     if (error.size())
         throwError(error);
 
-    createStream();
-    std::istream *stream(m_streamIf->m_istream);
+    // This will throw if the stream can't be opened.
+    LasStreamPtr lasStream = createStream();
+    std::istream& stream(*lasStream);
 
-    stream->seekg(0);
+    stream.seekg(0);
     // Always try to read as if we have 1.4 size.
     char headerBuf[las::Header::Size14];
-    stream->read(headerBuf, las::Header::Size14);
-    if (stream->gcount() < (std::streamsize)las::Header::Size12)
+    stream.read(headerBuf, las::Header::Size14);
+    if (stream.gcount() < (std::streamsize)las::Header::Size12)
         throwError("Couldn't read LAS header. File size insufficient.");
     d->header.fill(headerBuf, las::Header::Size14);
 
@@ -290,26 +333,36 @@ void LasReader::initializeLocal(PointTableRef table, MetadataNode& m)
     StringList errors = d->header.validate(fileSize, d->opts.nosrs);
     if (errors.size())
         throwError(errors.front());
-
     // Verify
     if (!las::pointFormatSupported(d->header.pointFormat()))
         throwError("Unsupported LAS input point format: " +
             Utils::toString((int)d->header.pointFormat()) + ".");
 
+    // Set the queue function based on whether we're compressed or not.
+    d->queueNext = d->header.dataCompressed() ?
+        std::bind(&LasReader::queueNextCompressedChunk, this) :
+        std::bind(&LasReader::queueNextStandardChunk, this);
+
+    // Set the load function based on whether we're 1.4 format or not.
+    using namespace std::placeholders;
+    d->loadPoint = d->header.has14PointFormat() ?
+        std::bind(&LasReader::loadPointV14, this, _1, _2, _3) :
+        std::bind(&LasReader::loadPointV10, this, _1, _2, _3);
+
     // Go peek into header and see if we are COPC
-    // Clear the error state since we potentially over-read the header, leaving
-    // the stream in error, when things are really fine for zero-point file.
-    stream->clear();
-    stream->seekg(377);
+    // If we over-read the file, the error state will be set, but things are really fine for
+    // a zero-point file, so clear the error.
+    stream.clear();
+    stream.seekg(377);
     char copcBuf[4] {};
-    stream->read(copcBuf, 4);
+    stream.read(copcBuf, 4);
     m.add("copc", ::memcmp(copcBuf, "copc", 4) == 0);
 
     // Read VLRs.
     // Clear the error state since the seek or read above may have failed but the file could
     // still be fine.
-    stream->clear();
-    stream->seekg(d->header.headerSize);
+    stream.clear();
+    stream.seekg(d->header.headerSize);
 
     char vlrHeaderBuf[las::Vlr::HeaderSize];
     std::vector<char> vlrBuf;
@@ -317,24 +370,38 @@ void LasReader::initializeLocal(PointTableRef table, MetadataNode& m)
     {
         las::Vlr vlr;
 
-        stream->read((char *)vlrHeaderBuf, las::Vlr::HeaderSize);
-        if (stream->gcount() != las::Vlr::HeaderSize)
-            throwError("Couldn't read VLR " + std::to_string(i + 1) + ". End of file reached.");
+        stream.read((char *)vlrHeaderBuf, las::Vlr::HeaderSize);
+        if (stream.gcount() != las::Vlr::HeaderSize)
+        {
+            if (d->opts.ignoreMissingVLRs)
+                break;
+            throwError("Couldn't read VLR " + std::to_string(i + 1) +
+                       ". End of file reached.");
+        }
         vlr.fillHeader(vlrHeaderBuf);
-        if ((uint64_t)stream->tellg() + vlr.promisedDataSize > d->header.pointOffset)
-            throwError("VLR " + std::to_string(i + 1) +
-                "(" + vlr.userId + "/" + std::to_string(vlr.recordId) + ") "
-                "size too large -- flows into point data.");
+        if ((uint64_t)stream.tellg() + vlr.promisedDataSize > d->header.pointOffset)
+        {
+            if (d->opts.ignoreMissingVLRs)
+                break;
+            throwError("VLR " + std::to_string(i + 1) + "(" + vlr.userId + "/" +
+                       std::to_string(vlr.recordId) + ") "
+                       "size too large -- flows into point data.");
+        }
         if (las::shouldIgnoreVlr(vlr, d->ignoreVlrs))
         {
-            stream->seekg(vlr.promisedDataSize, std::ios::cur);
+            stream.seekg(vlr.promisedDataSize, std::ios::cur);
             continue;
         }
         vlr.dataVec.resize(vlr.promisedDataSize);
-        stream->read(vlr.data(), vlr.promisedDataSize);
+        stream.read(vlr.data(), vlr.promisedDataSize);
 
-        if (stream->gcount() != (std::streamsize)vlr.promisedDataSize)
-            throwError("Couldn't read VLR " + std::to_string(i + 1) + ". End of file reached.");
+        if (stream.gcount() != (std::streamsize)vlr.promisedDataSize)
+        {
+            if (d->opts.ignoreMissingVLRs)
+                break;
+            throwError("Couldn't read VLR " + std::to_string(i + 1) +
+                       ". End of file reached.");
+        }
         d->vlrs.push_back(std::move(vlr));
     }
 
@@ -342,33 +409,33 @@ void LasReader::initializeLocal(PointTableRef table, MetadataNode& m)
     if (d->header.evlrOffset && d->header.evlrCount)
     {
         char evlrHeaderBuf[las::Evlr::HeaderSize];
-        stream->seekg(d->header.evlrOffset);
+        stream.seekg(d->header.evlrOffset);
         for (uint32_t i = 0; i < d->header.evlrCount; ++i)
         {
             las::Evlr evlr;
 
-            stream->read((char *)evlrHeaderBuf, las::Evlr::HeaderSize);
-            if (stream->gcount() != las::Evlr::HeaderSize)
+            stream.read((char *)evlrHeaderBuf, las::Evlr::HeaderSize);
+            if (stream.gcount() != las::Evlr::HeaderSize)
                 throwError("Couldn't read EVLR " + std::to_string(i + 1) +
                     ". End of file reached.");
             evlr.fillHeader(evlrHeaderBuf);
 
-            if ((uint64_t)stream->tellg() + evlr.promisedDataSize > fileSize)
+            if ((uint64_t)stream.tellg() + evlr.promisedDataSize > fileSize)
                 throwError("EVLR " + std::to_string(i + 1) +
                     "(" + evlr.userId + "/" + std::to_string(evlr.recordId) + ") "
                     "size too large -- exceeds file size.");
             if (las::shouldIgnoreVlr(evlr, d->ignoreVlrs))
             {
-                stream->seekg(evlr.promisedDataSize, std::ios::cur);
+                stream.seekg(evlr.promisedDataSize, std::ios::cur);
                 continue;
             }
             evlr.dataVec.resize(evlr.promisedDataSize);
-            stream->read(evlr.data(), evlr.promisedDataSize);
+            std::streampos pos = stream.tellg();
+            stream.read(evlr.data(), evlr.promisedDataSize);
 
-            //ABELL - Better error message.
-            if (stream->gcount() != (std::streamsize)evlr.promisedDataSize)
-                throwError("Couldn't read EVLR " + std::to_string(i + 1) +
-                    ". End of file reached.");
+            if (stream.gcount() != (std::streamsize)evlr.promisedDataSize)
+                throwError("Couldn't read EVLR " + std::to_string(i + 1) + " at offset " +
+                    std::to_string(pos) + ". Location is past the end of the file.");
             d->vlrs.push_back(std::move(evlr));
         }
     }
@@ -376,9 +443,20 @@ void LasReader::initializeLocal(PointTableRef table, MetadataNode& m)
     if (!d->opts.nosrs)
         d->srs.init(d->vlrs, d->opts.srsVlrOrder.types, d->header.mustUseWkt(), log());
 
-    if (d->opts.start > d->header.pointCount())
-        throwError("'start' value of " + std::to_string(d->opts.start) + " is too large. "
-            "File contains " + std::to_string(d->header.pointCount()) + " points.");
+    d->end = d->header.pointCount();
+    if (d->header.pointCount())
+    {
+        if (d->opts.start >= d->header.pointCount())
+            throwError("'start' value of " + std::to_string(d->opts.start) + " is too large. "
+                "File contains " + std::to_string(d->header.pointCount()) + " points.");
+
+        // maxPoints is positive because start is less than count from above.
+        uint64_t maxPoints = d->header.pointCount() - d->opts.start;
+
+        // count() can be a crazy-high value -- don't overflow with the addition.
+        if (count() < maxPoints)
+            d->end = d->opts.start + count();
+    }
 
     if (d->header.versionAtLeast(1, 4) || d->opts.useEbVlr)
         readExtraBytesVlr();
@@ -394,36 +472,154 @@ void LasReader::initializeLocal(PointTableRef table, MetadataNode& m)
 
 void LasReader::ready(PointTableRef table)
 {
-    if (m_count == 0) return;
+    if (getNumPoints() == 0)
+        return;
 
-    createStream();
-    std::istream *stream(m_streamIf->m_istream);
+    d->pool.resize(d->opts.numThreads);
+    LasStreamPtr lasStream(createStream());
+    std::istream& stream(*lasStream);
+
+    d->currentTile.reset();
+    d->tiles.clear();
 
     d->index = 0;
     if (d->header.dataCompressed())
     {
-        delete d->decompressor;
-
         const las::Vlr *vlr = las::findVlr(las::LaszipUserId, las::LaszipRecordId, d->vlrs);
         if (!vlr)
             throwError("LAZ file missing required laszip VLR.");
-        d->decompressor = new LazPerfVlrDecompressor(*stream, d->header, vlr->data());
+        lazperf::laz_vlr laz_vlr;
+        laz_vlr.fill(vlr->data(), vlr->dataSize());
+        try
+        {
+            d->chunkInfo.load(stream, d->header.pointOffset, d->header.pointCount(),
+                laz_vlr.chunk_size);
+        }
+        catch (const pdal_error& e)
+        {
+            throwError(e.what());
+        }
+
+        d->nextFetchChunk = 0;
+        d->nextFetchPoint = 0;
         if (d->opts.start > 0)
         {
-            if (d->opts.start > d->header.pointCount())
+            if (d->opts.start >= d->header.pointCount())
                 throwError("'start' option set past end of file.");
-            d->decompressor->seek(d->opts.start);
+            d->nextFetchChunk = d->chunkInfo.chunk(d->opts.start);
+            d->nextFetchPoint = d->chunkInfo.index(d->opts.start, d->nextFetchChunk);
         }
-        d->decompressorBuf.resize(d->header.pointSize);
+        d->nextReadChunk = d->nextFetchChunk;
     }
     else
     {
-        std::istream::pos_type start = d->header.pointOffset +
-            (d->opts.start * d->header.pointSize);
-        stream->seekg(start);
+        d->nextFetchPoint = d->opts.start;
+        d->nextFetchChunk = 0;
+        d->nextReadChunk = 0;
     }
+
+    for (int i = 0; i < d->opts.numThreads; ++i)
+        d->queueNext();
 }
 
+void LasReader::queueNextCompressedChunk()
+{
+    if ((d->nextFetchChunk >= d->chunkInfo.numChunks()) ||
+        (d->chunkInfo.firstPoint(d->nextFetchChunk) >= d->end))
+        return;
+
+    uint32_t chunk = d->nextFetchChunk;
+    uint32_t start = (uint32_t)d->nextFetchPoint;
+
+    d->pool.add([this, chunk, start]()
+    {
+        uint32_t chunkpoints = d->chunkInfo.chunkPoints(chunk);
+        uint64_t chunkoffset = d->chunkInfo.chunkOffset(chunk);
+        uint32_t chunksize = d->chunkInfo.chunkSize(chunk);
+
+        LasStreamPtr lasStream = createStream();
+        std::istream& in(*lasStream);
+
+        std::vector<char> buf(chunksize);
+        in.seekg(chunkoffset);
+        in.read(buf.data(), buf.size());
+
+        int32_t tilepoints = chunkpoints - start;
+        las::TilePtr tile = std::make_unique<las::Tile>(chunk, tilepoints * d->header.pointSize);
+
+        lazperf::reader::chunk_decompressor decomp(d->header.pointFormat(), d->header.ebCount(),
+            buf.data());
+
+        // We have to decompress all the points, even if we're discarding the points at
+        // the front because nextFetchPoint isn't 0. Just reuse the front of the tile
+        // buffer for discarded points.
+        char *pos = tile->data();
+        for (uint32_t i = 0; i < chunkpoints; ++i)
+        {
+            decomp.decompress(pos);
+
+            // Advance the point location in the tile if we're keeping the point.
+            if (i >= start)
+                pos += d->header.pointSize;
+        }
+        {
+            std::unique_lock l(d->mutex);
+            for (las::TilePtr& t : d->tiles)
+                if (!t)
+                {
+                    t = std::move(tile);
+                    goto done;
+                }
+            d->tiles.push_back(std::move(tile));
+        }
+        done:
+        d->processedCv.notify_one();
+    });
+    d->nextFetchChunk++;
+    // After the first chunk, we always start at 0.
+    d->nextFetchPoint = 0;
+}
+
+void LasReader::queueNextStandardChunk()
+{
+    const uint64_t chunkSize = 50'000;
+
+    if (d->nextFetchPoint >= d->end)
+        return;
+
+    int chunk = d->nextFetchChunk;
+    uint64_t start = d->nextFetchPoint;
+    uint64_t count = (std::min)(chunkSize, d->end - start);
+    d->pool.add([this, chunk, count, start]()
+    {
+        LasStreamPtr lasStream = createStream();
+        std::istream& in(*lasStream);
+
+        las::TilePtr tile = std::make_unique<las::Tile>(chunk, count * d->header.pointSize);
+        in.seekg(d->header.pointOffset + start * d->header.pointSize);
+        in.read(tile->data(), tile->size());
+
+        {
+            std::unique_lock l(d->mutex);
+            for (las::TilePtr& t : d->tiles)
+                if (!t)
+                {
+                    t = std::move(tile);
+                    goto done;
+                }
+            d->tiles.push_back(std::move(tile));
+        }
+        done:
+        d->processedCv.notify_one();
+    });
+
+    // This check is just to prevent overflow.
+    if (d->nextFetchPoint > (std::numeric_limits<uint64_t>::max)() - chunkSize)
+        d->nextFetchPoint = d->end;
+    else
+        d->nextFetchPoint += chunkSize;
+    d->nextFetchChunk++;
+}
 
 void LasReader::readExtraBytesVlr()
 {
@@ -481,112 +677,66 @@ void LasReader::addDimensions(PointLayoutPtr layout)
 
 bool LasReader::processOne(PointRef& point)
 {
-    if (d->index >= getNumPoints())
+    // This is called under lock. Note that we don't remove the tile *pointer* from the
+    // vector, it just gets set to null. When we add a tile, we'll look for a null
+    // entry before we add to the vector.
+    auto getTile = [this](uint32_t chunk)
+    {
+        for (las::TilePtr& t : d->tiles)
+            if (t && t->chunk() == chunk)
+                return std::move(t);
+        return las::TilePtr();
+    };
+
+    if (eof())
         return false;
 
-    if (d->header.dataCompressed())
+    // If we don't have an active tile, get the next one or wait for it to be ready.
+    if (!d->currentTile)
     {
-        if (!d->decompressor->decompress(d->decompressorBuf.data()))
-            throwError("Error reading point " + std::to_string(d->index) +
-                " from " + m_filename + ". Invalid/corrupt file.");
-        loadPoint(point, d->decompressorBuf.data(), d->header.pointSize);
-    }
-    else
-    {
-        std::vector<char> buf(d->header.pointSize);
-        m_streamIf->m_istream->read(buf.data(), buf.size());
-        loadPoint(point, buf.data(), buf.size());
+        {
+            std::unique_lock<std::mutex> l(d->mutex);
+            while (true)
+            {
+                d->currentTile = getTile(d->nextReadChunk);
+                if (d->currentTile)
+                    break;
+                d->processedCv.wait(l);
+            }
+        }
+
+        // Found the tile we wanted. Queue the next file read.
+        d->nextReadChunk++;
+        d->queueNext();
     }
+
+    // Load the point and advance the tile location.
+    d->loadPoint(point, d->currentTile->pos(), d->header.pointSize);
+    if (!d->currentTile->advance(d->header.pointSize))
+        d->currentTile.reset();
+
     d->index++;
     return true;
 }
 
-
 point_count_t LasReader::read(PointViewPtr view, point_count_t count)
 {
-    count = (std::min)(count, getNumPoints() - d->index);
+    count = (std::min)(count, getNumPoints() - (point_count_t)d->index);
 
     PointId i = 0;
-    if (d->header.dataCompressed())
+    for (i = 0; i < count; i++)
     {
-        for (i = 0; i < count; i++)
-        {
-            PointRef point = view->point(i);
-            PointId id = view->size();
-            processOne(point);
-            if (m_cb)
-                m_cb(*view, id);
-        }
+        PointRef point = view->point(i);
+        PointId id = view->size();
+        processOne(point);
+        if (m_cb)
+            m_cb(*view, id);
     }
-    else
-    {
-        point_count_t remaining = count;
-
-        // Make a buffer at most a meg.
-        size_t bufsize = (std::min)((point_count_t)1000000, count * d->header.pointSize);
-        std::vector<char> buf(bufsize);
-        try
-        {
-            do
-            {
-                point_count_t blockPoints = readFileBlock(buf, remaining);
-                remaining -= blockPoints;
-                char *pos = buf.data();
-                while (blockPoints--)
-                {
-                    PointId id = view->size();
-                    PointRef point = view->point(id);
-                    loadPoint(point, pos, d->header.pointSize);
-                    if (m_cb)
-                        m_cb(*view, id);
-                    pos += d->header.pointSize;
-                    i++;
-                }
-            } while (remaining);
-        }
-        catch (std::out_of_range&)
-        {}
-        catch (invalid_stream&)
-        {}
-    }
-    d->index += i;
     return (point_count_t)i;
 }
 
 
-point_count_t LasReader::readFileBlock(std::vector<char>& buf, point_count_t maxpoints)
-{
-    std::istream *stream(m_streamIf->m_istream);
-
-    size_t ptLen = d->header.pointSize;
-    point_count_t blockpoints = buf.size() / ptLen;
-
-    blockpoints = (std::min)(maxpoints, blockpoints);
-    if (stream->eof())
-        throw invalid_stream("stream is done");
-
-    stream->read(buf.data(), blockpoints * ptLen);
-    if (stream->gcount() != (std::streamsize)(blockpoints * ptLen))
-    {
-        // we read fewer bytes than we asked for
-        // because the file was either truncated
-        // or the header is bunk.
-        blockpoints = stream->gcount() / ptLen;
-    }
-    return blockpoints;
-}
-
-
-void LasReader::loadPoint(PointRef& point, char *buf, size_t bufsize)
-{
-    if (d->header.has14PointFormat())
-        loadPointV14(point, buf, bufsize);
-    else
-        loadPointV10(point, buf, bufsize);
-}
-
-
-void LasReader::loadPointV10(PointRef& point, char *buf, size_t bufsize)
+void LasReader::loadPointV10(PointRef& point, const char *buf, size_t bufsize)
 {
     LeExtractor istream(buf, bufsize);
 
@@ -667,7 +817,7 @@ void LasReader::loadPointV10(PointRef& point, char *buf, size_t bufsize)
 }
 
 
-void LasReader::loadPointV14(PointRef& point, char *buf, size_t bufsize)
+void LasReader::loadPointV14(PointRef& point, const char *buf, size_t bufsize)
 {
     LeExtractor istream(buf, bufsize);
 
@@ -770,12 +920,15 @@ void LasReader::loadExtraDims(LeExtractor& istream, PointRef& point)
 
 void LasReader::done(PointTableRef)
 {
-    m_streamIf.reset();
+    d->pool.join();
+    if (d->isRemote)
+        FileUtils::deleteFile(m_filename);
 }
 
 bool LasReader::eof()
 {
-    return d->index >= getNumPoints();
+    // This breaks when the number of points is the maximum (2^64 - 1), but that's never happening.
+    return d->index >= d->end;
 }
 
 
